import { Text } from '@codemirror/text';
export { Text } from '@codemirror/text';

declare enum MapMode {
    Simple = 0,
    TrackDel = 1,
    TrackBefore = 2,
    TrackAfter = 3
}
declare class ChangeDesc {
    get length(): number;
    get newLength(): number;
    get empty(): boolean;
    iterGaps(f: (posA: number, posB: number, length: number) => void): void;
    iterChangedRanges(f: (fromA: number, toA: number, fromB: number, toB: number) => void, individual?: boolean): void;
    get invertedDesc(): ChangeDesc;
    composeDesc(other: ChangeDesc): ChangeDesc;
    mapDesc(other: ChangeDesc, before?: boolean): ChangeDesc;
    mapPos(pos: number, assoc?: number): number;
    mapPos(pos: number, assoc: number, mode: MapMode): number | null;
    touchesRange(from: number, to?: number): boolean | "cover";
}
declare type ChangeSpec = {
    from: number;
    to?: number;
    insert?: string | Text;
} | ChangeSet | readonly ChangeSpec[];
declare class ChangeSet extends ChangeDesc {
    apply(doc: Text): Text;
    mapDesc(other: ChangeDesc, before?: boolean): ChangeDesc;
    invert(doc: Text): ChangeSet;
    compose(other: ChangeSet): ChangeSet;
    map(other: ChangeDesc, before?: boolean): ChangeSet;
    iterChanges(f: (fromA: number, toA: number, fromB: number, toB: number, inserted: Text) => void, individual?: boolean): void;
    get desc(): ChangeDesc;
    toJSON(): any;
    static of(changes: ChangeSpec, length: number, lineSep?: string): ChangeSet;
    static empty(length: number): ChangeSet;
    static fromJSON(json: any): ChangeSet;
}

declare class SelectionRange {
    readonly from: number;
    readonly to: number;
    private flags;
    get anchor(): number;
    get head(): number;
    get empty(): boolean;
    get assoc(): -1 | 0 | 1;
    get bidiLevel(): number | null;
    get goalColumn(): number | undefined;
    map(change: ChangeDesc, assoc?: number): SelectionRange;
    extend(from: number, to?: number): SelectionRange;
    eq(other: SelectionRange): boolean;
    toJSON(): any;
    static fromJSON(json: any): SelectionRange;
}
declare class EditorSelection {
    readonly ranges: readonly SelectionRange[];
    readonly mainIndex: number;
    map(change: ChangeDesc, assoc?: number): EditorSelection;
    eq(other: EditorSelection): boolean;
    get main(): SelectionRange;
    asSingle(): EditorSelection;
    addRange(range: SelectionRange, main?: boolean): EditorSelection;
    replaceRange(range: SelectionRange, which?: number): EditorSelection;
    toJSON(): any;
    static fromJSON(json: any): EditorSelection;
    static single(anchor: number, head?: number): EditorSelection;
    static create(ranges: readonly SelectionRange[], mainIndex?: number): EditorSelection;
    static cursor(pos: number, assoc?: number, bidiLevel?: number, goalColumn?: number): SelectionRange;
    static range(anchor: number, head: number, goalColumn?: number): SelectionRange;
}

declare type FacetConfig<Input, Output> = {
    combine?: (value: readonly Input[]) => Output;
    compare?: (a: Output, b: Output) => boolean;
    compareInput?: (a: Input, b: Input) => boolean;
    static?: boolean;
    enables?: Extension;
};
declare class Facet<Input, Output = readonly Input[]> {
    private isStatic;
    private constructor();
    static define<Input, Output = readonly Input[]>(config?: FacetConfig<Input, Output>): Facet<Input, Output>;
    of(value: Input): Extension;
    compute(deps: readonly Slot<any>[], get: (state: EditorState) => Input): Extension;
    computeN(deps: readonly Slot<any>[], get: (state: EditorState) => readonly Input[]): Extension;
    from(field: StateField<Input>): Extension;
    from<T>(field: StateField<T>, get: (value: T) => Input): Extension;
}
declare type Slot<T> = Facet<any, T> | StateField<T> | "doc" | "selection";
declare type StateFieldSpec<Value> = {
    create: (state: EditorState) => Value;
    update: (value: Value, transaction: Transaction) => Value;
    compare?: (a: Value, b: Value) => boolean;
    provide?: (field: StateField<Value>) => Extension;
    toJSON?: (value: Value, state: EditorState) => any;
    fromJSON?: (json: any, state: EditorState) => Value;
};
declare class StateField<Value> {
    private createF;
    private updateF;
    private compareF;
    private constructor();
    static define<Value>(config: StateFieldSpec<Value>): StateField<Value>;
    private create;
    init(create: (state: EditorState) => Value): Extension;
    extension: Extension;
}
declare type Extension = {
    extension: Extension;
} | readonly Extension[];
declare const Prec: {
    fallback: (ext: Extension) => Extension;
    default: (ext: Extension) => Extension;
    extend: (ext: Extension) => Extension;
    override: (ext: Extension) => Extension;
};
declare function tagExtension(tag: string | symbol, extension: Extension): Extension;

declare class Annotation<T> {
    readonly type: AnnotationType<T>;
    readonly value: T;
    static define<T>(): AnnotationType<T>;
    private _isAnnotation;
}
declare class AnnotationType<T> {
    of(value: T): Annotation<T>;
}
interface StateEffectSpec<Value> {
    map?: (value: Value, mapping: ChangeDesc) => Value | undefined;
}
declare class StateEffect<Value> {
    readonly value: Value;
    map(mapping: ChangeDesc): StateEffect<Value> | undefined;
    is<T>(type: StateEffectType<T>): this is StateEffect<T>;
    static define<Value = null>(spec?: StateEffectSpec<Value>): StateEffectType<Value>;
    static mapEffects(effects: readonly StateEffect<any>[], mapping: ChangeDesc): readonly StateEffect<any>[];
}
declare class StateEffectType<Value> {
    readonly map: (value: any, mapping: ChangeDesc) => any | undefined;
    of(value: Value): StateEffect<Value>;
}
interface TransactionSpec {
    changes?: ChangeSpec;
    selection?: EditorSelection | {
        anchor: number;
        head?: number;
    };
    effects?: StateEffect<any> | readonly StateEffect<any>[];
    annotations?: Annotation<any> | readonly Annotation<any>[];
    scrollIntoView?: boolean;
    reconfigure?: ReconfigurationSpec;
    filter?: boolean;
    sequential?: boolean;
}
interface ReconfigurationSpec {
    full?: Extension;
    append?: Extension;
    [tag: string]: Extension | undefined;
}
declare class Transaction {
    readonly startState: EditorState;
    readonly changes: ChangeSet;
    readonly selection: EditorSelection | undefined;
    readonly effects: readonly StateEffect<any>[];
    readonly reconfigure: ReconfigurationSpec | undefined;
    readonly scrollIntoView: boolean;
    get newDoc(): Text;
    get newSelection(): EditorSelection;
    get state(): EditorState;
    annotation<T>(type: AnnotationType<T>): T | undefined;
    get docChanged(): boolean;
    static time: AnnotationType<number>;
    static userEvent: AnnotationType<string>;
    static addToHistory: AnnotationType<boolean>;
}

declare enum CharCategory {
    Word = 0,
    Space = 1,
    Other = 2
}

interface EditorStateConfig {
    doc?: string | Text;
    selection?: EditorSelection | {
        anchor: number;
        head?: number;
    };
    extensions?: Extension;
}
declare class EditorState {
    readonly doc: Text;
    readonly selection: EditorSelection;
    field<T>(field: StateField<T>): T;
    field<T>(field: StateField<T>, require: false): T | undefined;
    update(...specs: readonly TransactionSpec[]): Transaction;
    replaceSelection(text: string | Text): {
        changes: ChangeSet;
        selection: EditorSelection;
        effects: readonly StateEffect<any>[];
    };
    changeByRange(f: (range: SelectionRange) => {
        range: SelectionRange;
        changes?: ChangeSpec;
        effects?: StateEffect<any> | readonly StateEffect<any>[];
    }): {
        changes: ChangeSet;
        selection: EditorSelection;
        effects: readonly StateEffect<any>[];
    };
    changes(spec?: ChangeSpec): ChangeSet;
    toText(string: string): Text;
    sliceDoc(from?: number, to?: number): string;
    facet<Output>(facet: Facet<any, Output>): Output;
    toJSON(fields?: {
        [prop: string]: StateField<any>;
    }): any;
    static fromJSON(json: any, config?: EditorStateConfig, fields?: {
        [prop: string]: StateField<any>;
    }): EditorState;
    static create(config?: EditorStateConfig): EditorState;
    static allowMultipleSelections: Facet<boolean, boolean>;
    static tabSize: Facet<number, number>;
    get tabSize(): number;
    static lineSeparator: Facet<string, string | undefined>;
    get lineBreak(): string;
    static phrases: Facet<{
        [key: string]: string;
    }, readonly {
        [key: string]: string;
    }[]>;
    phrase(phrase: string): string;
    static languageData: Facet<(state: EditorState, pos: number) => readonly {
        [name: string]: any;
    }[], readonly ((state: EditorState, pos: number) => readonly {
        [name: string]: any;
    }[])[]>;
    languageDataAt<T>(name: string, pos: number): readonly T[];
    charCategorizer(at: number): (char: string) => CharCategory;
    static changeFilter: Facet<(tr: Transaction) => boolean | readonly number[], readonly ((tr: Transaction) => boolean | readonly number[])[]>;
    static transactionFilter: Facet<(tr: Transaction) => TransactionSpec | readonly TransactionSpec[], readonly ((tr: Transaction) => TransactionSpec | readonly TransactionSpec[])[]>;
    static transactionExtender: Facet<(tr: Transaction) => Pick<TransactionSpec, "effects" | "annotations" | "reconfigure"> | null, readonly ((tr: Transaction) => Pick<TransactionSpec, "effects" | "annotations" | "reconfigure"> | null)[]>;
}

declare type StateCommand = (target: {
    state: EditorState;
    dispatch: (transaction: Transaction) => void;
}) => boolean;

declare function combineConfig<Config>(configs: readonly Partial<Config>[], defaults: Partial<Config>, // Should hold only the optional properties of Config, but I haven't managed to express that
combine?: {
    [P in keyof Config]?: (first: Config[P], second: Config[P]) => Config[P];
}): Config;

export { Annotation, AnnotationType, ChangeDesc, ChangeSet, ChangeSpec, CharCategory, EditorSelection, EditorState, EditorStateConfig, Extension, Facet, MapMode, Prec, ReconfigurationSpec, SelectionRange, StateCommand, StateEffect, StateEffectType, StateField, Transaction, TransactionSpec, combineConfig, tagExtension };
